/*
Copyright (C) 2004 Jacquelin POTIER <jacquelin.potier@free.fr>
Dynamic aspect ratio code Copyright (C) 2004 Jacquelin POTIER <jacquelin.potier@free.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//-----------------------------------------------------------------------------
// Object: provides a Link list
//-----------------------------------------------------------------------------

#include "linklist.h"

CLinkList::CLinkList(size_t ItemSize)
{
    this->ItemSize=ItemSize;
    this->Head=NULL;
    this->Tail=NULL;
    this->hevtListUnlocked=CreateEvent(NULL,FALSE,TRUE,NULL);
    this->ItemsNumber=0;
}

CLinkList::~CLinkList(void)
{
    // remove all items
    this->RemoveAllItems();

    // close unlocked event
    if (this->hevtListUnlocked)
    {
        CloseHandle(this->hevtListUnlocked);
        this->hevtListUnlocked=NULL;
    }
}

//-----------------------------------------------------------------------------
// Name: AddItem
// Object: Add Item to the link list allocating necessary memory for ItemData
// Parameters :
//     in  : 
//     out :
//     return : CLinkListItem* Item : pointer to the added item
//-----------------------------------------------------------------------------
CLinkListItem* CLinkList::AddItem()
{
    CLinkListItem* Item;
    
    WaitForSingleObject(this->hevtListUnlocked,INFINITE);
    Item = new CLinkListItem();
    if (Item==NULL)
    {
        SetEvent(this->hevtListUnlocked);
        return NULL;
    }

    // allocate memory for item data 
    Item->ItemData = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, this->ItemSize);

    if (this->Head == NULL)
    {
        this->Head = Item;
        this->Tail = Item;
    }
    else
    {
        this->Tail->NextItem = Item;
        this->Tail = Item;
    }  

    this->Tail->NextItem = NULL;

    this->ItemsNumber++;
    SetEvent(this->hevtListUnlocked);
    return Item;
}

//-----------------------------------------------------------------------------
// Name: AddItem
// Object: Add Item to the link list allocating necessary memory for ItemData
// Parameters :
//     in  : PVOID PointerToItemData : pointer to data to add
//     out :
//     return : CLinkListItem* Item : pointer to the added item
//-----------------------------------------------------------------------------
CLinkListItem* CLinkList::AddItem(PVOID PointerToItemData)
{
    CLinkListItem* pLinkListItem=this->AddItem();
    if (pLinkListItem==NULL)
        return NULL;
    // copy data information
    memcpy(pLinkListItem->ItemData,PointerToItemData,this->ItemSize);
    return pLinkListItem;
}


//-----------------------------------------------------------------------------
// Name: RemoveItem
// Object: Remove item from list and free memory pointed by ItemData
// Parameters :
//     in  : CLinkListItem* Item : pointer to the item to remove
//     out :
//     return : 
//-----------------------------------------------------------------------------
void CLinkList::RemoveItem(CLinkListItem* Item)
{
    WaitForSingleObject(this->hevtListUnlocked,INFINITE);

    CLinkListItem* CurrentItem = this->Head;
    if (this->Head == NULL)
    {
        SetEvent(this->hevtListUnlocked);
        return;
    }
    if (Item == this->Head)
    {
        if (this->Head == this->Tail) 
        {
            // free item data memory
            HeapFree(GetProcessHeap(), 0, Item->ItemData);
            Item->ItemData=NULL;
            Item->NextItem=NULL;
            // free item itself
            delete Item;

            this->Head = NULL;// must be done after freeing item
        }
        else
        {
            this->Head = this->Head->NextItem;// must be done before freeing item

            // free item data memory
            HeapFree(GetProcessHeap(), 0, Item->ItemData);
            Item->ItemData=NULL;
            Item->NextItem=NULL;
            // free item itself
            delete Item;
            Item=NULL;
        }

        this->ItemsNumber--;
        SetEvent(this->hevtListUnlocked);
        return;
    }

    while ((CurrentItem->NextItem != Item)&&(CurrentItem->NextItem != NULL))
        CurrentItem = CurrentItem->NextItem;

    // if item not found
    if (CurrentItem->NextItem==NULL)
    {
        SetEvent(this->hevtListUnlocked);
        return;
    }

    CurrentItem->NextItem = Item->NextItem;

    if (Item == this->Tail)
        this->Tail = CurrentItem;

    // free item data memory
    HeapFree(GetProcessHeap(), 0, Item->ItemData);
    Item->ItemData=NULL;
    // free item itself
    delete Item;

    this->ItemsNumber--;
    SetEvent(this->hevtListUnlocked);
}

//-----------------------------------------------------------------------------
// Name: RemoveItemFromItemData
// Object: Remove an item identify by its data
//          Data should be unique in list, else first matching item will be removed
// Parameters :
//     in  : PointerToItemData : a pointer to Item data
//                               check is done for content of data
//     out :
//     return : 
//-----------------------------------------------------------------------------
void CLinkList::RemoveItemFromItemData(PVOID PointerToItemData)
{
    if (IsBadReadPtr(PointerToItemData,this->ItemSize))
        return;

    WaitForSingleObject(this->hevtListUnlocked,INFINITE);

    CLinkListItem* CurrentItem = this->Head;
    CLinkListItem* Item;
    if (this->Head == NULL)
    {
        SetEvent(this->hevtListUnlocked);
        return;
    }
    if (memcmp(PointerToItemData,this->Head->ItemData,this->ItemSize)==0)
    {
        Item=this->Head;
        if (this->Head == this->Tail) 
        {
            // free item data memory
            HeapFree(GetProcessHeap(), 0, Item->ItemData);
            Item->ItemData=NULL;
            Item->NextItem=NULL;
            // free item itself
            delete Item;

            this->Head = NULL;// must be done after freeing item
        }
        else
        {
            this->Head = this->Head->NextItem;// must be done before freeing item

            // free item data memory
            HeapFree(GetProcessHeap(), 0, Item->ItemData);
            Item->ItemData=NULL;
            Item->NextItem=NULL;
            // free item itself
            delete Item;
            Item=NULL;
        }

        this->ItemsNumber--;
        SetEvent(this->hevtListUnlocked);
        return;
    }

    while (CurrentItem->NextItem != NULL)
    {
        if (memcmp(CurrentItem->NextItem->ItemData,PointerToItemData,this->ItemSize)==0)
            break;
        CurrentItem = CurrentItem->NextItem;
    }

    // if item not found
    if (CurrentItem->NextItem==NULL)
    {
        SetEvent(this->hevtListUnlocked);
        return;
    }
    Item=CurrentItem->NextItem;
    CurrentItem->NextItem = Item->NextItem;

    if (Item == this->Tail)
        this->Tail = CurrentItem;

    // free item data memory
    HeapFree(GetProcessHeap(), 0, Item->ItemData);
    Item->ItemData=NULL;
    // free item itself
    delete Item;

    this->ItemsNumber--;
    SetEvent(this->hevtListUnlocked);
}

//-----------------------------------------------------------------------------
// Name: RemoveAllItems
// Object: Remove all items from the list
// Parameters :
//     in  : 
//     out :
//     return : 
//-----------------------------------------------------------------------------
void CLinkList::RemoveAllItems()
{
    while (this->Head)
        RemoveItem(this->Head);
}

//-----------------------------------------------------------------------------
// Name: GetItemsCount
// Object: return the number of items in the list
// Parameters :
//     in  : 
//     out :
//     return : number of items
//-----------------------------------------------------------------------------
DWORD CLinkList::GetItemsCount()
{
    return this->ItemsNumber;
}

//-----------------------------------------------------------------------------
// Name: ToInsecureArray
// Object: return an array of pointer to elements (that means if you 
//           store an object 'obj' you'll get an obj** as result)
//         Speedest way but items are no more accessible as soon as they
//          are remove from list
//         CALLER HAVE TO FREE ARRAY calling "delete[] Array;"
// Parameters :
//     in  : 
//     out : DWORD* pdwArraySize : returned array size
//     return : array of pointer to elements, NULL on error or if no elements
//-----------------------------------------------------------------------------
PVOID* CLinkList::ToInsecureArray(DWORD* pdwArraySize)
{
    PVOID* Array;
    DWORD dwCounter;

    // check param
    if (IsBadWritePtr(pdwArraySize,sizeof(DWORD)))
        return NULL;

    // wait for synchro
    WaitForSingleObject(this->hevtListUnlocked,INFINITE);

    // get number of Items
    *pdwArraySize=this->GetItemsCount();
    
    if (*pdwArraySize == 0)
    {
        SetEvent(this->hevtListUnlocked);
        return NULL;
    }

    // we just allocate an array to pointer list
    // this is the speedest way as we don't need 
    // to do memory allocation for each Item
    Array=new PVOID[*pdwArraySize];
    CLinkListItem* CurrentItem = this->Head;
    for (dwCounter=0;dwCounter<*pdwArraySize;dwCounter++)// as while (CurrentItem->NextItem != NULL)) but more secure
    {
        Array[dwCounter]=CurrentItem->ItemData;
        CurrentItem = CurrentItem->NextItem;
    }
    SetEvent(this->hevtListUnlocked);
    return Array;
}

//-----------------------------------------------------------------------------
// Name: ToSecureArray
// Object: return an array of elements (that means if you 
//           store an object 'obj' you'll get an obj* as result)
//         Items are accessible even if they are removed from list and list object is destroyed
//         CALLER HAVE TO FREE ARRAY calling "delete[] Array;"
// Parameters :
//     in  : 
//     out : DWORD* pdwArraySize : returned array size
//     return : array of elements, NULL on error or if no elements
//-----------------------------------------------------------------------------
PVOID CLinkList::ToSecureArray(DWORD* pdwArraySize)
{
    BYTE* Array;
    DWORD dwCounter;

    // check param
    if (IsBadWritePtr(pdwArraySize,sizeof(DWORD)))
        return NULL;

    // wait for synchro
    WaitForSingleObject(this->hevtListUnlocked,INFINITE);

    // get number of Items
    *pdwArraySize=this->GetItemsCount();
    
    if (*pdwArraySize == 0)
    {
        SetEvent(this->hevtListUnlocked);
        return NULL;
    }

    // we just allocate an array to pointer list
    // this is the speedest way as we don't need 
    // to do memory allocation for each Item
    Array=new BYTE[(*pdwArraySize)*this->ItemSize];
    CLinkListItem* CurrentItem = this->Head;
    for (dwCounter=0;dwCounter<*pdwArraySize;dwCounter++)// as while (CurrentItem->NextItem != NULL)) but more secure
    {
        memcpy(&Array[dwCounter*this->ItemSize],CurrentItem->ItemData,this->ItemSize);
        CurrentItem = CurrentItem->NextItem;
    }
    SetEvent(this->hevtListUnlocked);
    return (PVOID)Array;
}