/*
Copyright (C) 2004 Jacquelin POTIER <jacquelin.potier@free.fr>
Dynamic aspect ratio code Copyright (C) 2004 Jacquelin POTIER <jacquelin.potier@free.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//-----------------------------------------------------------------------------
// Object: PE helper
//-----------------------------------------------------------------------------

#include "pe.h"

CPE::CPE(TCHAR* filename)
{
    _tcsncpy(this->pcFilename,filename,MAX_PATH);
    memset(&this->DosHeader,0,sizeof(IMAGE_DOS_HEADER));
    this->pSectionHeaders=NULL;
}

CPE::~CPE(void)
{
    if (this->pSectionHeaders)
        delete this->pSectionHeaders;
}
void CPE::ShowError(TCHAR* pcMsg)
{
    MessageBox(NULL,pcMsg,_T("Error"),MB_OK|MB_ICONERROR|MB_TOPMOST);
}

//-----------------------------------------------------------------------------
// Name: Parse
// Object: loads DosHeader, NTHeader and pSectionHeaders infos from file
// Parameters :
//     in  : 
//     out :
//     return : FALSE on error
//-----------------------------------------------------------------------------
BOOL CPE::Parse()
{
    BOOL bRes;
    HANDLE hFile;
    DWORD dwSize;
    DWORD dwReadSize=0;

    // open file
    hFile = CreateFile(this->pcFilename, GENERIC_READ, FILE_SHARE_READ, NULL,
                        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
                    
    if (hFile==INVALID_HANDLE_VALUE)
    {   
        CAPIError::ShowLastError();
        return FALSE;
    }
    // Try to obtain hFile's size 
    dwSize = GetFileSize (hFile, NULL) ; 
    // If we failed ... 
    if (dwSize == INVALID_FILE_SIZE) 
    {
        CAPIError::ShowLastError();
        CloseHandle(hFile);
        return FALSE;
    }
    this->pBuffer=new unsigned char[dwSize];
    if (!ReadFile(hFile,this->pBuffer,dwSize,&dwReadSize,NULL))
    {
        CAPIError::ShowLastError();
        delete this->pBuffer;
        CloseHandle(hFile);
        return FALSE;
    }

    // check for exe
    if (((PIMAGE_DOS_HEADER)this->pBuffer)->e_magic ==IMAGE_DOS_SIGNATURE)
    {
        // store DOS header
        memcpy(&this->DosHeader,this->pBuffer,sizeof(IMAGE_DOS_HEADER));
        // parse NTHeader (and the remaining of the file)
        bRes=this->ParseIMAGE_NT_HEADERS();
        if (bRes)
            bRes=this->ParseIMAGE_SECTION_HEADER();
    }
    else
    {
        this->ShowError(_T("Unrecognized file format."));
        bRes=FALSE;
    }

    delete this->pBuffer;
    CloseHandle(hFile);

    return bRes;
}

//-----------------------------------------------------------------------------
// Name: ParseIMAGE_NT_HEADERS
// Object: loads NTHeader infos
// Parameters :
//     in  : 
//     out :
//     return : FALSE on error
//-----------------------------------------------------------------------------
BOOL CPE::ParseIMAGE_NT_HEADERS()
{
	PIMAGE_NT_HEADERS pNTHeader;
	
	pNTHeader = (PIMAGE_NT_HEADERS)(this->pBuffer+this->DosHeader.e_lfanew);
	// First, verify that the e_lfanew field gave us a reasonable
	// pointer, then verify the PE signature.
	if (IsBadReadPtr(pNTHeader, sizeof(IMAGE_NT_HEADERS)) ||
	     pNTHeader->Signature != IMAGE_NT_SIGNATURE )
	{
        this->ShowError(_T("Error reading IMAGE_NT_HEADERS"));
		return FALSE;
	}
    // store NTHeader
    memcpy(&this->NTHeader,pNTHeader,sizeof(IMAGE_NT_HEADERS));

    return TRUE;
}
//-----------------------------------------------------------------------------
// Name: ParseIMAGE_SECTION_HEADER
// Object: loads pSectionHeaders infos
// Parameters :
//     in  : 
//     out :
//     return : FALSE on error
//-----------------------------------------------------------------------------
BOOL CPE::ParseIMAGE_SECTION_HEADER()
{
    PIMAGE_SECTION_HEADER pSecHeader;
    // get sections informations
    pSecHeader=(PIMAGE_SECTION_HEADER)(this->pBuffer+this->DosHeader.e_lfanew+sizeof(IMAGE_NT_HEADERS));
    if (IsBadReadPtr(pSecHeader, sizeof(IMAGE_SECTION_HEADER)*this->NTHeader.FileHeader.NumberOfSections))
	{
        this->ShowError(_T("Error in Sections Header"));
		return FALSE;
	}
    // free previous section headers if any
    if (this->pSectionHeaders!=NULL)
        delete this->pSectionHeaders;
    // allocate memory and copy data
    this->pSectionHeaders=new IMAGE_SECTION_HEADER[this->NTHeader.FileHeader.NumberOfSections];
    // get all sections infos
    memcpy(this->pSectionHeaders,pSecHeader,sizeof(IMAGE_SECTION_HEADER)*this->NTHeader.FileHeader.NumberOfSections);

    return TRUE;
}

//-----------------------------------------------------------------------------
// Name: SaveIMAGE_DOS_HEADER
// Object: save DosHeader to file
// Parameters :
//     in  : 
//     out :
//     return : FALSE on error
//-----------------------------------------------------------------------------
BOOL CPE::SaveIMAGE_DOS_HEADER()
{
    HANDLE hFile;
    BOOL bRet=TRUE;
    DWORD dwNbBytesWritten;
    // open file
    hFile = CreateFile(this->pcFilename, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL,
                        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (hFile==INVALID_HANDLE_VALUE)
    {
        CAPIError::ShowLastError();
        return FALSE;
    }
    // write IMAGE_DOS_HEADER
    if (!WriteFile(hFile,&this->DosHeader,sizeof(IMAGE_DOS_HEADER),&dwNbBytesWritten,NULL))
    {
        CAPIError::ShowLastError();
        bRet=FALSE;
    }
    // close file
    CloseHandle(hFile);
    return bRet;
}

//-----------------------------------------------------------------------------
// Name: SaveIMAGE_NT_HEADERS
// Object: save NTHeader to file
// Parameters :
//     in  : 
//     out :
//     return : FALSE on error
//-----------------------------------------------------------------------------
BOOL CPE::SaveIMAGE_NT_HEADERS()
{
    HANDLE hFile;
    BOOL bRet=TRUE;
    DWORD dwNbBytesWritten;
    // open file
    hFile = CreateFile(this->pcFilename, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL,
                        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (hFile==INVALID_HANDLE_VALUE)
    {
        CAPIError::ShowLastError();
        return FALSE;
    }
    // move to IMAGE_NT_HEADERS start
    SetFilePointer(hFile,this->DosHeader.e_lfanew,0,FILE_CURRENT);
    // write IMAGE_NT_HEADERS
    if (!WriteFile(hFile,&this->NTHeader,sizeof(IMAGE_NT_HEADERS),&dwNbBytesWritten,NULL))
    {
        CAPIError::ShowLastError();
        bRet=FALSE;
    }
    
    // close file
    CloseHandle(hFile);
    return bRet;
}

//-----------------------------------------------------------------------------
// Name: SavePIMAGE_SECTION_HEADER
// Object: save PIMAGE_SECTION_HEADER to file
//         WARNING currently this func don't allow you to change the number of 
//         sections. It's just allow you to modify fields of theses sections.
//         If number of sections is changed you'll probably get an unworking EXE
// Parameters :
//     in  : 
//     out :
//     return : FALSE on error
//-----------------------------------------------------------------------------
BOOL CPE::SavePIMAGE_SECTION_HEADER()
{
    HANDLE hFile;
    BOOL bRet=TRUE;
    DWORD dwNbBytesWritten;
    // open file
    hFile = CreateFile(this->pcFilename, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL,
                        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (hFile==INVALID_HANDLE_VALUE)
    {
        CAPIError::ShowLastError();
        return FALSE;
    }
    // move to IMAGE_NT_HEADERS start
    SetFilePointer(hFile,this->DosHeader.e_lfanew+sizeof(IMAGE_NT_HEADERS),0,FILE_CURRENT);

    // write IMAGE_SECTION_HEADER array
    if (!WriteFile(hFile,this->pSectionHeaders,sizeof(IMAGE_SECTION_HEADER)*this->NTHeader.FileHeader.NumberOfSections,&dwNbBytesWritten,NULL))
    {
        CAPIError::ShowLastError();
        bRet=FALSE;
    }
    
    // close file
    CloseHandle(hFile);
    return bRet;
}


//-----------------------------------------------------------------------------
// Name: GetChecksum
// Object: return the current file checksum stored in NTHeader.OptionalHeader.CheckSum
//         and the real file checksum
// Parameters :
//     in  : 
//     out :
//     return : FALSE on error
//-----------------------------------------------------------------------------
BOOL CPE::GetChecksum(DWORD* pCurrentHeaderCheckSum,DWORD* pRealCheckSum)
{
    HANDLE hFile;
    HANDLE hFileMapping;
    LPVOID BaseAddress;

    if (IsBadWritePtr(pCurrentHeaderCheckSum,sizeof(DWORD))
        || IsBadWritePtr(pRealCheckSum,sizeof(DWORD)))
        return FALSE;

    // open file
    hFile = CreateFile(this->pcFilename, GENERIC_READ, FILE_SHARE_READ, NULL,
                        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (hFile==INVALID_HANDLE_VALUE)
    {
        CAPIError::ShowLastError();
        return FALSE;
    }
    // create file mapping
    hFileMapping=CreateFileMapping(hFile,NULL,PAGE_READONLY,0,0,NULL);
    if (!hFileMapping)
    {
        CAPIError::ShowLastError();
        CloseHandle(hFile);
        return FALSE;
    }
    // map view of file
    BaseAddress=MapViewOfFile(hFileMapping,FILE_MAP_READ,0,0,0);

    if (BaseAddress==NULL)
    {
        CAPIError::ShowLastError();
        CloseHandle(hFileMapping);
        CloseHandle(hFile);
        return FALSE;
    }

    // get checksum infos
    if (!CheckSumMappedFile(
                        BaseAddress,
                        GetFileSize(hFile,NULL), // FileLength
                        pCurrentHeaderCheckSum,  // HeaderSum
                        pRealCheckSum            // CheckSum
                        )
       )
    if (!hFileMapping)
    {
        CAPIError::ShowLastError();
        UnmapViewOfFile(BaseAddress);
        CloseHandle(hFileMapping);
        CloseHandle(hFile);
        return FALSE;
    }

    // unmap view of file
    UnmapViewOfFile(BaseAddress);
    // close file mapping
    CloseHandle(hFileMapping);
    // close file
    CloseHandle(hFile);
    return TRUE;
}
//-----------------------------------------------------------------------------
// Name: CorrectChecksum
// Object: comput and corredt the checksum in NTHeader.OptionalHeader.CheckSum
//         Warning this func make a parsing before writting checksum
//         --> DosHeader, NTHeader and pSectionHeaders information will match 
//             file content and modifications to these member will be lost
//             exept if you save them before calling this func
// Parameters :
//     in  : 
//     out :
//     return : FALSE on error
//-----------------------------------------------------------------------------
BOOL CPE::CorrectChecksum()
{
    DWORD dwCurrentCheckSum;
    DWORD dwRealCheckSum;
    if (!this->Parse())
        return FALSE;
    if (!this->GetChecksum(&dwCurrentCheckSum,&dwRealCheckSum))
        return FALSE;
    this->NTHeader.OptionalHeader.CheckSum=dwRealCheckSum;

    return this->SaveIMAGE_NT_HEADERS();
}

//-----------------------------------------------------------------------------
// Name: RawToRva
// Object: Translate raw address to relative virtual one
// Parameters :
//     in  : DWORD RawAddress
//     out : DWORD* pRvaAddress
//     return : FALSE on error
//-----------------------------------------------------------------------------
BOOL CPE::RawToRva(DWORD RawAddress,DWORD* pRvaAddress)
{
    TCHAR pszMsg[MAX_PATH];
    if (IsBadWritePtr(pRvaAddress,sizeof(DWORD)))
        return FALSE;
    for (int cnt=0;cnt<this->NTHeader.FileHeader.NumberOfSections;cnt++)
    {
        if ((this->pSectionHeaders[cnt].PointerToRawData<RawAddress)
            && (this->pSectionHeaders[cnt].PointerToRawData+this->pSectionHeaders[cnt].SizeOfRawData>RawAddress))
        {
            *pRvaAddress=RawAddress-this->pSectionHeaders[cnt].PointerToRawData+this->pSectionHeaders[cnt].VirtualAddress;
            return TRUE;
        }
    }
    _stprintf(pszMsg,_T("Relative Virtual Adress not found for Raw address %.8X"),RawAddress);
    MessageBox(NULL,pszMsg,_T("Error"),MB_OK|MB_ICONERROR|MB_TOPMOST);  
    return FALSE;
}

//-----------------------------------------------------------------------------
// Name: RawToRva
// Object: Translate relative virtual address to raw one
// Parameters :
//     in  : DWORD RvaAddress
//     out : DWORD* pRawAddress
//     return : FALSE on error
//-----------------------------------------------------------------------------
BOOL CPE::RvaToRaw(DWORD RvaAddress,DWORD* pRawAddress)
{
    TCHAR pszMsg[MAX_PATH];
    if (IsBadWritePtr(pRawAddress,sizeof(DWORD)))
        return FALSE;
    for (int cnt=0;cnt<this->NTHeader.FileHeader.NumberOfSections;cnt++)
    {
        if ((this->pSectionHeaders[cnt].VirtualAddress<RvaAddress)
            && (this->pSectionHeaders[cnt].VirtualAddress+this->pSectionHeaders[cnt].SizeOfRawData>RvaAddress))
        {
            *pRawAddress=RvaAddress-this->pSectionHeaders[cnt].VirtualAddress+this->pSectionHeaders[cnt].PointerToRawData;
            return TRUE;
        }
    }
    _stprintf(pszMsg,_T("Raw address not found for Relative Virtual Adress %.8X"),RvaAddress);
    MessageBox(NULL,pszMsg,_T("Error"),MB_OK|MB_ICONERROR|MB_TOPMOST);  
    return FALSE;
}