/*
Copyright (C) 2004 Jacquelin POTIER <jacquelin.potier@free.fr>
Dynamic aspect ratio code Copyright (C) 2004 Jacquelin POTIER <jacquelin.potier@free.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//-----------------------------------------------------------------------------
// Object: class helper for using a Mailslot server
//-----------------------------------------------------------------------------

#include "mailslotserver.h"

//-----------------------------------------------------------------------------
// Name: CMailSlotServer
// Object: Constructor
// Parameters :
//     in  : - TCHAR* pszMailSlotName : slot name
//                  \\.\mailslot\name Retrieves a client handle to a local mailslot. 
//                  \\computername\mailslot\name Retrieves a client handle to a remote mailslot. 
//                  \\domainname\mailslot\name Retrieves a client handle to all mailslots with the specified name in the specified domain. 
//                  \\*\mailslot\name Retrieves a client handle to all mailslots with the specified name in the system's primary domain. 
//           - DWORD dwMaxMessageSize : max message size
//           - pfnMailSlotCallback CallbackFunc : pointer to a callback function taking in parameter 
//                                 PVOID pData : pointer to receive data
//                                 PVOID pUserData : user data = what you want
//           - PVOID pUserData : user data given back at each callback can be used to store pointer of object
//                               as call back is not object oriented
//     out : 
//     return : 
//-----------------------------------------------------------------------------
CMailSlotServer::CMailSlotServer(TCHAR* pszMailSlotName,pfnMailSlotCallback CallbackFunc,PVOID pUserData)
{

    _tcsncpy(this->pszMailSlotName,pszMailSlotName,MAX_PATH);
    this->pszMailSlotName[MAX_PATH-1]=0;
    
    this->CallbackFunc=CallbackFunc;
    this->pUserData=pUserData;

    this->bStarted=FALSE;
    this->hMailslot=NULL;
    this->hThreadHandle=NULL;

    // create events
    this->hevtStopEvent=CreateEvent(NULL,FALSE,FALSE,NULL);
}

CMailSlotServer::~CMailSlotServer(void)
{
    // make sure thread is stopped
    this->Stop();
    CleanCloseHandle(&this->hevtStopEvent);
}
//-----------------------------------------------------------------------------
// Name: Start
// Object: Start mailslot server to use it
// Parameters :
//     in  :
//     out :
//     return : TRUE on success, FALSE on error
//-----------------------------------------------------------------------------
BOOL CMailSlotServer::Start(void)
{
    if (this->bStarted)
        this->Stop();

    // assume we have a callback
    if (IsBadCodePtr((FARPROC)this->CallbackFunc))
        return FALSE;

    // create mailslot
    this->hMailslot=CreateMailslot(this->pszMailSlotName,0,MAILSLOT_WAIT_FOREVER,0);
    if (hMailslot == INVALID_HANDLE_VALUE)
        return FALSE;

    // reset events
    ResetEvent(this->hevtStopEvent);

    // create a listening thread
    this->hThreadHandle=CreateThread(NULL,0,CMailSlotServerThreadListener,this,0,NULL);
    if (!this->hThreadHandle)
        return FALSE;

    // set bStarted flag
    this->bStarted=TRUE;

    return TRUE;
}
//-----------------------------------------------------------------------------
// Name: Stop
// Object: Stop mailslot server
// Parameters :
//     in  :
//     out :
//     return : TRUE on success, FALSE on error
//-----------------------------------------------------------------------------
BOOL CMailSlotServer::Stop(void)
{
    if (!this->bStarted)
        return TRUE;
    // send stop event
    SetEvent(this->hevtStopEvent);
    this->CallbackFunc=NULL;
    // wait for end of thread
    WaitForSingleObject(this->hThreadHandle,CMAILSLOTSERVER_CLOSINGSERVER_MAX_WAIT);

    //destroy mailslot
    CleanCloseHandle(&this->hMailslot);
    // close thread handle
    CleanCloseHandle(&this->hThreadHandle);
    // set bStarted flag
    this->bStarted=FALSE;
    return TRUE;
}

//-----------------------------------------------------------------------------
// Name: CMailSlotServerThreadListener
// Object: Get messages and call user call back. Stop on event hevtStopEvent
// Parameters :
//     in  : LPVOID lpParam : pointer to a CMailSlotServer object
//     out :
//     return : 0 on success, 0xFFFFFFFF on error
//-----------------------------------------------------------------------------
DWORD WINAPI CMailSlotServerThreadListener(LPVOID lpParam)
{
    DWORD dwResult;
    DWORD dwError;
    DWORD dwBytesRead=0;
    DWORD dwMsgSize=0;
    DWORD dwMessageCount;

    CMailSlotServer* pCMSS=(CMailSlotServer*)lpParam;
    DWORD dwMaxMessageSize=512;
    BYTE* pData=new BYTE[dwMaxMessageSize];
    OVERLAPPED Overlapped;
    memset(&Overlapped,0,sizeof(OVERLAPPED));
    Overlapped.hEvent=CreateEvent(NULL,FALSE,FALSE,NULL);
    HANDLE ph[2]={pCMSS->hevtStopEvent,Overlapped.hEvent};
    for (;;)
    {
        // see CMailSlotClient : in the same write operation we write data size on first DWORD
        // and next data
        if (!ReadFile(pCMSS->hMailslot, pData, dwMsgSize, &dwBytesRead, &Overlapped))
        {
            dwError=GetLastError();
            if (dwError==ERROR_INSUFFICIENT_BUFFER)
            {
                if(GetMailslotInfo(pCMSS->hMailslot, 0, &dwMsgSize, &dwMessageCount, NULL))
                {
                    // adjust our local buffer size
                    if (dwMaxMessageSize<dwMsgSize)
                    {
                        delete pData;
                        pData=new BYTE[dwMsgSize];
                        dwMaxMessageSize=dwMsgSize;
                    }
                    // restart the read operation with correct buffer size
                    continue;
                }
                else
                {
                    CAPIError::ShowError(dwError);
                    delete pData;
                    return dwError;
                }
            }

            if (dwError!=ERROR_IO_PENDING)
            {
                CAPIError::ShowError(dwError);
                delete pData;
                return dwError;
            }
        }
        // wait for overlapped event or stop event
        dwResult = WaitForMultipleObjects(2,ph,FALSE,INFINITE);
        switch (dwResult)
        {
        case WAIT_OBJECT_0:// stop event
            CleanCloseHandle(&Overlapped.hEvent);
            delete pData;
            return 0;
        case WAIT_OBJECT_0+1:// read event
            if (dwMsgSize==0)
                break;
            if (!IsBadCodePtr((FARPROC)pCMSS->CallbackFunc))
                pCMSS->CallbackFunc(pData,dwBytesRead,pCMSS->pUserData);
            // gives a new fake buffer size to allow to catch ERROR_INSUFFICIENT_BUFFER error
            dwMsgSize=0;
            break;

        case WAIT_FAILED:
            CleanCloseHandle(&Overlapped.hEvent);
            delete pData;
            return WAIT_FAILED;
        }
    }
}
